@@ -29,6 +29,14 @@
 #define FALLOC_FL_PUNCH_HOLE	0x02
 #endif
 
+#ifndef FALLOC_FL_COLLAPSE_RANGE
+#define FALLOC_FL_COLLAPSE_RANGE 0x08
+#endif
+
+#ifndef FALLOC_FL_ZERO_RANGE
+#define FALLOC_FL_ZERO_RANGE 0x10
+#endif
+
 static cmdinfo_t allocsp_cmd;
 static cmdinfo_t freesp_cmd;
 static cmdinfo_t resvsp_cmd;
@@ -37,6 +45,8 @@ static cmdinfo_t zero_cmd;
 #if defined(HAVE_FALLOCATE)
 static cmdinfo_t falloc_cmd;
 static cmdinfo_t fpunch_cmd;
+static cmdinfo_t fcollapse_cmd;
+static cmdinfo_t fzero_cmd;
 #endif
 
 static int
@@ -159,8 +169,11 @@ fallocate_f(
 	int		mode = 0;
 	int		c;
 
-	while ((c = getopt(argc, argv, "kp")) != EOF) {
+	while ((c = getopt(argc, argv, "ckp")) != EOF) {
 		switch (c) {
+		case 'c':
+			mode = FALLOC_FL_COLLAPSE_RANGE;
+			break;
 		case 'k':
 			mode = FALLOC_FL_KEEP_SIZE;
 			break;
@@ -203,6 +216,50 @@ fpunch_f(
 	}
 	return 0;
 }
+
+static int
+fcollapse_f(
+	int		argc,
+	char		**argv)
+{
+	xfs_flock64_t	segment;
+	int		mode = FALLOC_FL_COLLAPSE_RANGE;
+
+	if (!offset_length(argv[1], argv[2], &segment))
+		return 0;
+
+	if (fallocate(file->fd, mode,
+			segment.l_start, segment.l_len)) {
+		perror("fallocate");
+		return 0;
+	}
+	return 0;
+}
+
+static int
+fzero_f(
+	int		argc,
+	char		**argv)
+{
+	xfs_flock64_t	segment;
+	int		mode = FALLOC_FL_ZERO_RANGE;
+	int		index = 1;
+
+	if (strncmp(argv[index], "-k", 3) == 0) {
+		mode |= FALLOC_FL_KEEP_SIZE;
+		index++;
+	}
+
+	if (!offset_length(argv[index], argv[index + 1], &segment))
+		return 0;
+
+	if (fallocate(file->fd, mode,
+			segment.l_start, segment.l_len)) {
+		perror("fallocate");
+		return 0;
+	}
+	return 0;
+}
 #endif	/* HAVE_FALLOCATE */
 
 void
@@ -263,9 +320,9 @@ prealloc_init(void)
 	falloc_cmd.argmin = 2;
 	falloc_cmd.argmax = -1;
 	falloc_cmd.flags = CMD_NOMAP_OK | CMD_FOREIGN_OK;
-	falloc_cmd.args = _("[-k] [-p] off len");
+	falloc_cmd.args = _("[-c] [-k] [-p] off len");
 	falloc_cmd.oneline =
-		_("allocates space associated with part of a file via fallocate");
+	_("allocates space associated with part of a file via fallocate");
 	add_command(&falloc_cmd);
 
 	fpunch_cmd.name = "fpunch";
@@ -275,7 +332,27 @@ prealloc_init(void)
 	fpunch_cmd.flags = CMD_NOMAP_OK | CMD_FOREIGN_OK;
 	fpunch_cmd.args = _("off len");
 	fpunch_cmd.oneline =
-		_("de-allocates space assocated with part of a file via fallocate");
+	_("de-allocates space assocated with part of a file via fallocate");
 	add_command(&fpunch_cmd);
+
+	fcollapse_cmd.name = "fcollapse";
+	fcollapse_cmd.cfunc = fcollapse_f;
+	fcollapse_cmd.argmin = 2;
+	fcollapse_cmd.argmax = 2;
+	fcollapse_cmd.flags = CMD_NOMAP_OK | CMD_FOREIGN_OK;
+	fcollapse_cmd.args = _("off len");
+	fcollapse_cmd.oneline =
+	_("de-allocates space and eliminates the hole by shifting extents");
+	add_command(&fcollapse_cmd);
+
+	fzero_cmd.name = "fzero";
+	fzero_cmd.cfunc = fzero_f;
+	fzero_cmd.argmin = 2;
+	fzero_cmd.argmax = 3;
+	fzero_cmd.flags = CMD_NOMAP_OK | CMD_FOREIGN_OK;
+	fzero_cmd.args = _("[-k] off len");
+	fzero_cmd.oneline =
+	_("zeroes space and eliminates holes by preallocating");
+	add_command(&fzero_cmd);
 #endif	/* HAVE_FALLOCATE */
 }
